var util = require("util");

/*
  Generic functions for NodeJS.

  TODO:
    - Should intrinsic types be dispatched like their object-based counterparts?
      I.e., should 'string' be dispatched as a 'String' and vice-versa?
    - Method dispatch examines leaf object types, but not the prototype
      hierarchy. Walk the prototype hierarchy to find the matching method.
    - Support 0-arg generic functions.
 */

/**
  If true, will print debug messages.
 */
var DEBUG = false;
var debug = (DEBUG
    ? function () { util.log.apply(util.log, arguments); }
    : function () { });

function error(reason) {
  throw new Error(reason);
};

/**
  Since {arguments} is not a regular array and thus has none of the methods of
  an array but can be handled like an array one under the hood, we use
  {Array.prototype}'s functions through {call}.
 */
var slice = Array.prototype.slice;
var reduce = Array.prototype.reduce;

/**
  Typeof is insufficient for our needs. We need better control over what type a
  value evaluates to.
  @param x {any} A value.
  @return {string} A type identifier for the given value.
 */
function type(x) {
  var t = typeof x;
  switch (t) {
    /*
       None of the values in any recent JavaScript implementations have a "null"
       type, but let's be prepared for the future.
     */
    case "null":
    case "undefined":
    case "boolean":
    case "number":
    case "string":
      break;
    case "object":
      if (x === null) {
        /*
           Normally (typeof null) -> "object", but that's wrong for our
           purposes.  We don't want a method that dispatches on "object" to
           dispatch on "null". We fix it here.
         */
        t = "null";
      }
			else if (Array.isArray(x)) {
				t = "array";
			}
      else if (x.constructor.name == "Object") {
        /*
           (typeof {}) -> "object", even though ({}).constructor.name ->
           "Object". We combine those two to simply "object".
         */
        t = "object";
      }
      else {
        t = x.constructor.name;
      }
      break;
    default:
      error("Unhandled type: "+t);
  }
  return t;
};


/**
  @param x {any} A value.
	@param type {string} A type identifier.
  @return {boolean} Whether or not the value {x} is an instance of {type}.
 */
function typep(x, type) {
	return type(x) === type;
}

/**
  @param x {any} A value.
	@return {array} The types {x} is an instance of.
 */
function types(x) {
	function iter(x, list) {
		if (x === undefined) {
			return list;
		}
		else {
			list.push(x.name);
			return iter(x.super_, list);
		}
	};
	return iter(x.constructor, []);
};

/**
  @param list {array} A list of values.
  @return {array} A list of types.
 */
function typeList(list) {
  return list.map(function (x) {
		return type(x);
	});
};

/**
  Creates a new instance of GenericFunction.
  @param numParameters {number} The number of parameters the function takes.
  @throws {Error} If numParameters is negative.
 */
function GenericFunction(numParameters) {
  if (numParameters < 0) {
    error("numParameters must be positive");
  }

  var self = this;
  var methods = {};

  function signature(list) {
    return list.toString();
  };

  function lookupMethod(sig) {
    return methods[sig];
  };

  function storeMethod(sig, met) {
    methods[sig] = met;
  };

  function deleteMethod(sig) {
    delete methods[sig];
  };

  /**
    Adds a new method to the generic function.
    @param list {Array} The type signature for the method.
    @param met {Function} The function to invoke if the type signature matches
    the supplied one.
    @throws {Error} If an incorrect number of parameters have been supplied.
   */
  this.addMethod = function (list, met) {
    if (list.length !== numParameters) {
      error("You must supply exactly "+numParameters+" arguments.");
    }
    var sig = signature(list);
    if (lookupMethod(sig) === undefined) {
      debug("Adding method with signature "+sig+" to "+numParameters+"-param method.");
    }
    else {
      debug("Replacing method with signature "+sig+".");
    }
    storeMethod(list, met);
    return met;
  };

  /**
    Finds a method associated with this generic function.
    @param list {Array} The type signature of the method to find.
    @return {null,Function} If a method was found, its {Function} object. If a
    method wasn't found, then {null}.
   */
  this.findMethod = function (list) {
    if (list.length !== numParameters) {
      /*
         We shouldn't throw an exception if the user passes in an incorrect
         number of parameters. A method for that signature simply won't exist.
         It's her fault.
       */
      return null;
    }
    var sig = signature(list);
    var met = lookupMethod(sig);
    return (met === undefined ? null : met);
  };

  this.removeMethod = function (list) {
    if (list.length !== numParameters) {
      error("You must supply exactly "+numParameters+" arguments.");
    }
    var sig = signature(list);
    var met = lookupMethod(sig);
    if (met !== undefined) {
      deleteMethod(sig);
      return true;
    }
    else {
      return false;
    }
  };

  /**
    The generic function's equivalent to "call".
    @param arguments[0..len-1] {any} The parameters to the generic function.
    @throws {Error} If the number of parameters doesn't match the generic
    function's number of parameters.
   */
  this.dispatch = function (args) {
    if (args.length !== numParameters) {
      error("Invalid number of arguments: "+args.length+". Required: "+numParameters+".");
    }
    var list = typeList(slice.call(args, 0, args.length));
    var sig = signature(list);
    var met = lookupMethod(sig);
    if (met === undefined) {
      error("No applicable method for parameter list "+list);
    }
    return met.apply(met, args);
  };

  debug("Created new "+numParameters+"-param generic function.");
};

/**
  @param numParameters {number} The number of parameters the generic function
  will accept.
  @return {Function} A callable function that will dispatch on the types of its
  parameters.
 */
function generic(numParameters /*, rest */) {

  var gen = new GenericFunction(numParameters);
  var functor = function () {
    return gen.dispatch(arguments);
  };
 
  /**
    @param arguments[0..len-2] The type signature.
    @param arguments[len-1] The function.
    @return {Function} The {Function} that was added to the generic function.
   */
  functor.addMethod = function () {
    var list = slice.call(arguments, 0, arguments.length-1);
    var met = arguments[arguments.length-1];
    return gen.addMethod(list, met);
  };

  /**
    @param arguments[0..len-1] The type signature.
    @return {Function} The {Function} that was added to the generic function.
   */
  functor.findMethod = function () {
    var list = slice.call(arguments, 0, arguments.length);
    return gen.findMethod(list);
  };

  functor.removeMethod = function () {
    var list = slice.call(arguments, 0, arguments.length);
    return gen.removeMethod(list);
  };

  return functor;
};

module.exports = generic;
