/**
	Test suite for generic functions.
	@author Chris Eineke
	@package generic-function-test
	@version 0.0.1
 */
var util = require("util");
var generic = require("../lib/generic-function");

exports["0-arg: unsupported"] = function (test) {
	test.throws(function () { generic(0); }, Error);
	test.done();
};

exports["1-arg: no methods -> throws exception"] = function (test) {
	var gen = generic(1);
	test.throws(function () { gen(""); }, Error);
	test.done();
};

exports["1-arg: 1 method: specialized for number"] = function (test) {
	var gen = generic(1);
	var met = function (x) { return 42; };
	gen.addMethod("number", met);
	test.equal(gen.findMethod("number"), met);
	test.equal(gen(0), 42);
	test.equal(gen.removeMethod("number"), true);
	test.equal(gen.findMethod("number"), null);
	test.done();
};

exports["1-arg: 1 method: specialized for string"] = function (test) {
	var gen = generic(1);
	var met = function (x) { return "foo"; }
	gen.addMethod("string", met);
	test.equal(gen.findMethod("string", met));
	test.equal(gen(""), "foo");
	test.equal(gen.removeMethod("string"), true);
	test.equal(gen.findMethod("string"), null);
	test.done();
};

exports["1-arg: 1 method: specialized for object"] = function (test) {
	var gen = generic(1);
	var met = function (x) { return { "foo": 42 }; };
	gen.addMethod("object", met);
	test.equal(gen.findMethod("object", met));
	test.equal(gen({}).toString(), { "foo": 42 }.toString());
	test.equal(gen.removeMethod("object"), true);
	test.equal(gen.findMethod("object"), null);
	test.done();
};

exports["1-arg: 3 methods: specialized for number, string, and object"] = function (test) {
	var gen = generic(1);
	var met1 = function (x) { return 42; };
	var met2 = function (x) { return "foo"; };
	var met3 = function (x) { return { "foo": 42 }; };
	gen.addMethod("number", met1);
	gen.addMethod("string", met2);
	gen.addMethod("object", met3);
	test.equal(gen.findMethod("number", met1));
	test.equal(gen.findMethod("string", met2));
	test.equal(gen.findMethod("object", met3));
	test.equal(gen(1), 42);
	test.equal(gen(""), "foo");
	test.equal(gen({}).toString(), { "foo": 42 }.toString());
	test.equal(gen.removeMethod("number"), true);
	test.equal(gen.removeMethod("string"), true);
	test.equal(gen.removeMethod("object"), true);
	test.equal(gen.findMethod("number"), null);
	test.equal(gen.findMethod("string"), null);
	test.equal(gen.findMethod("object"), null);
	test.done();
};

exports["1-args: all intrinsic types"] = function (test) {
	var gen = generic(1);
	var met1 = function (x) { return "number"; };
	var met2 = function (x) { return "string"; };
	var met3 = function (x) { return "object"; };
	var met4 = function (x) { return "null"; };
	var met5 = function (x) { return "undefined"; };
	gen.addMethod("number", met1);
	gen.addMethod("string", met2);
	gen.addMethod("object", met3);
	gen.addMethod("null", met4);
	gen.addMethod("undefined", met5);
	test.equal(gen.findMethod("number", met1));
	test.equal(gen.findMethod("string", met2));
	test.equal(gen.findMethod("object", met3));
	test.equal(gen.findMethod("null", met4));
	test.equal(gen.findMethod("undefined", met5));
	var data = [ 1, "foo", { }, null, undefined ];
	var result = data.map(function (x) { return gen(x); });
	var expected = "number,string,object,null,undefined";
	test.equal(result.toString(), expected);
	test.equal(gen.removeMethod("number"), true);
	test.equal(gen.removeMethod("string"), true);
	test.equal(gen.removeMethod("object"), true);
	test.equal(gen.removeMethod("null"), true);
	test.equal(gen.removeMethod("undefined"), true);
	test.equal(gen.findMethod("number"), null);
	test.equal(gen.findMethod("string"), null);
	test.equal(gen.findMethod("object"), null);
	test.equal(gen.findMethod("null"), null);
	test.equal(gen.findMethod("undefined"), null);
	test.done();
};

exports["2-arg: no methods -> throws exception"] = function (test) {
	var gen = generic(2);
	test.throws(function () { gen("foo"); }, Error);
	test.done();
};

exports["2-arg: 4 methods"] = function (test) {
	var gen = generic(2);
	var met1 = function (theFirst, theSecond) { return 2; };
	var met2 = function (theString, theNumber) { return 4; };
	var met3 = function (theString, theNumber) { return 8; };
	var met4 = function (theFirst, theSecond) { return 16; };
	gen.addMethod("number", "number", met1);
	gen.addMethod("number", "string", met2);
	gen.addMethod("string", "number", met3);
	gen.addMethod("string", "string", met4);
	test.equal(gen.findMethod("number", "number"), met1);
	test.equal(gen.findMethod("number", "string"), met2);
	test.equal(gen.findMethod("string", "number"), met3);
	test.equal(gen.findMethod("string", "string"), met4);
	test.equal(gen(1, 1), 2);
	test.equal(gen(1, "foo"), 4);
	test.equal(gen("bar", 1), 8);
	test.equal(gen("foo", "bar"), 16);
	test.equal(gen.removeMethod("number", "number"), true);
	test.equal(gen.removeMethod("number", "string"), true);
	test.equal(gen.removeMethod("string", "number"), true);
	test.equal(gen.removeMethod("string", "string"), true);
	test.equal(gen.findMethod("number", "number"), null);
	test.equal(gen.findMethod("number", "string"), null);
	test.equal(gen.findMethod("string", "number"), null);
	test.equal(gen.findMethod("string", "string"), null);
	test.done();
};

exports["bank account example"] = function (test) {
	function BankAccount(balance) {
		this.balance = balance;
	};
	function CheckingAccount(balance) {
		this.constructor.super_.call(this, balance);
	};
	util.inherits(CheckingAccount, BankAccount);
	function SavingsAccount(balance) {
		this.constructor.super_.call(this, balance);
	};
	util.inherits(SavingsAccount, BankAccount);

	var balanceOf = generic(1);
	balanceOf.addMethod("BankAccount", function (account) { return account.balance; });
	balanceOf.addMethod("CheckingAccount", function (account) { return account.balance; });
	balanceOf.addMethod("SavingsAccount", function (account) { return account.balance; });

	var withdraw = generic(2);
	withdraw.addMethod("BankAccount", "number", function (account, amount) {
		if (account.balance < amount) {
			throw new Error("Account overdrawn.");
		}
		account.balance -= amount;
	});
	withdraw.addMethod("CheckingAccount", "number", function (account, amount) {
		account.balance -= amount;
	});
	withdraw.addMethod("SavingsAccount", "number", function (account, amount) {
		throw new Error("This account allows only deposits.");
	});

	var acc1 = new BankAccount(100);
	withdraw(acc1, 80);
	test.equal(balanceOf(acc1), 20);

	var acc2 = new CheckingAccount(100);
	withdraw(acc2, 150);
	test.equal(balanceOf(acc2), -50);

	var acc3 = new SavingsAccount(100);
	test.throws(function () { withdraw(acc3, 1); }, Error);
	test.equal(balanceOf(acc3), 100);

	test.done();
};

exports["shape example"] = function (test) {
	function Shape() { }

	function Circle(radius) {
		this.constructor.super_.call(this);
		this.radius = radius;
	};
	util.inherits(Circle, Shape);

	function Rectangle(width, height) {
		this.constructor.super_.call(this);
		this.width = width;
		this.height = height;
	};
	util.inherits(Rectangle, Shape);

	var area = generic(1);
	area.addMethod("Shape", function (shape) {
		throw new Error("Cannot calculate area of abstract shape.");
	});
	area.addMethod("Circle", function (circle) {
		return 2 * Math.PI * circle.radius * circle.radius;
	});
	area.addMethod("Rectangle", function (rectangle) {
		return rectangle.width * rectangle.height;
	});

	var shape = new Shape();
	var circle = new Circle(1);
	var rectangle = new Rectangle(2, 2);

	test.throws(function () { area(shape); }, Error);
	test.equal(area(circle), 6.283185307179586);
	test.equal(area(rectangle), 4);

	test.done();
};

exports["simple performance test"] = function (test) {
	function Base() { };
	function Derived(value) {
		this.constructor.super_.call(this);
		this.value = value;
	};
	util.inherits(Derived, Base);

	var instances = [];
	for (var i=0, N=1000000 ; i<N ; i++) {
		instances.push(new Derived(i));
	}

	var accu = 0;
	var valueOf = generic(1);
	valueOf.addMethod("Derived", function (derived) {
		return derived.value * derived.value;
	});
	var before = Date.now();
	for (var i=0, N=1000000 ; i<N ; i++) {
		accu += valueOf(instances[i]);
	}
	var after = Date.now();
	console.log(after-before);

	test.done();
};

exports["simple performance test (without generic)"] = function (test) {
	function Base() {
		this.getValue = null;
	};
	function Derived(value) {
		this.constructor.super_.call(this);
		this.getValue = function () { return this.value * this.value; };
		this.value = value;
	};
	util.inherits(Derived, Base);

	var instances = [];
	for (var i=0, N=1000000 ; i<N ; i++) {
		instances.push(new Derived(i));
	}

	var accu = 0;
	var valueOf = generic(1);
	var before = Date.now();
	for (var i=0, N=1000000 ; i<N ; i++) {
		accu += instances[i].getValue();
	}
	var after = Date.now();
	console.log(after-before);

	test.done();
};
